/*
 *	rieblmac: display/set MAC address in Riebl VME board
 *
 *	usage:	rieblmac [mac address]
 *
 *	When invoked without arguments, rieblmac displays the MAC address(es)
 *	currently stored in the Riebl board: the one in the 'init block'
 *	which is used by the lance hardware, and the one in the configuration
 *	(reserved) area as used by the original Atari Network Software.  The
 *	presence of the latter address is signalled by a magic number, and may
 *	not be present.
 *
 *	In order to change the MAC address, you may supply the new one as an
 *	argument in the format 00:00:36:04:nn:nn, where nn is any hexadecimal
 *	value, and ':' represents a separator which may be ':', '.', or '-'.
 *	Note that the new MAC address is constrained to begin with 00:00:36:04
 *	since this is the range assigned to Riebl VME boards.
 *
 *	Also note that no effort is made to check for the presence of a Riebl
 *	board; however, since only the TT and the Mega STe have VME slots, an
 *	error message is issued if the program is run on other machines.
 *
 *
 *	Remarks on the Riebl board, packet buffers, and memory addresses
 *	----------------------------------------------------------------
 *	The Riebl board has 65536 bytes of memory.  The LANCE architecture
 *	uses the first 24 bytes as the init block, and the last 2 bytes
 *	are used to hold the interrupt vector number, leaving 65510 bytes
 *	for packet buffers.
 *
 *	Each packet buffer requires 1518 bytes for data and 8 bytes for
 *	the message descriptor entry, for a total of 1526 bytes.  So the
 *	maximum number of packet buffers is 65510/1526 = 42.  However,
 *	the number of packet buffers for receive and transmit must each be
 *	a power of 2, so the maximum number of buffers that can be specified
 *	is 40 (32 receive and 8 transmit, or vice versa).
 *
 *	Assuming 32 receive & 8 transmit, the relative memory addresses for
 *	the various structures is as follows:
 *		0x0000		init block
 *		0x0018		8 transmit message descriptors
 *		0x0058		32 receive message descriptors
 *		0x0158		40 packet buffers
 *		0xee88		unused
 *
 *	The original Atari Network Software package reserved 0xee70-0xeec0,
 *	so part of this is overwriten by this arrangement.  However, the
 *	only part of this reserved area that anyone else even slightly cares
 *	about is a magic number at 0xee8a, and the MAC address at 0xee8e,
 *	both of which are preserved with this setup.
 *
 *
 *	Version history
 *	---------------
 *	v1.0	december/2007	roger burrows
 *		supports Riebl VME ethernet board on TT or Mega STe
 *		(Mega STe as yet untested)
 *
 */
#include <stdio.h>
#include <stdlib.h>

#include <cookie.h>
#include <osbind.h>
#include <string.h>

#define MCH_MEGASTE		0x00010010L
#define MCH_TT030		0x00020000L

#define BASE_MEGASTE	(char *)0x00c10000L
#define BASE_TT030		(char *)0xfe010000L

#define RIEBL_MAGIC		0x09051990L			/* ANS (Atari Network Software) compatibility */
#define MAGIC_OFFSET	0xee8a
#define HWADDR_OFFSET	0xee8e

#define INIT_OFFSET		2					/* Lance architecture ... */

#define MAC_PREFIX		0x00003604L			/* first 4 bytes of valid MAC address for Riebl */

/* prototypes */
int atox(char *hex);
int copy(void);
void display_mac(char *desc,char *mac);
void quit(int rc,char *msg);
int valid_mac(char *mac,char *s);
void xfer_riebl(char *to,char *from,int length);

int main(int argc,char **argv)
{
int error;
long mchtype, magic;
char *base;
char init_mac[6], config_mac[6], updt_mac[6];
char swab_mac[6];							/* byte-swapped */

	if (getcookie(_MCH,&mchtype) == 0)
		quit(1,"Can't determine machine type");

	if (mchtype == MCH_MEGASTE)
		base = BASE_MEGASTE;
	else if (mchtype == MCH_TT030)
		base = BASE_TT030;
	else quit(1,"Not a TT/Mega STe");

	xfer_riebl((char *)&magic,base+MAGIC_OFFSET,4);
	if (magic == RIEBL_MAGIC)
		xfer_riebl(config_mac,base+HWADDR_OFFSET,6);
	xfer_riebl(init_mac,base+INIT_OFFSET,6);
	swab(init_mac,init_mac,6);

	if (magic == RIEBL_MAGIC)
		display_mac("Configured",config_mac);
	else printf("There is no configured MAC address\n");
	display_mac("Init block",init_mac);

	if (argc < 2)
		quit(0,"");
	printf("\n");

	if (!valid_mac(updt_mac,argv[1]))
		quit(1,"Invalid new MAC address (must start with 00:00:36:04)");

	/*
	 *	update (and see if it worked)
	 */
	error = 0;
	magic = RIEBL_MAGIC;					/* first we update the reserved area */
	xfer_riebl(base+MAGIC_OFFSET,(char *)&magic,4);
	xfer_riebl(base+HWADDR_OFFSET,updt_mac,6);
	xfer_riebl(config_mac,base+HWADDR_OFFSET,6);
	if (memcmp(updt_mac,config_mac,6) != 0)
		error++;

	swab(updt_mac,swab_mac,6);				/* swab for init block */
	xfer_riebl(base+INIT_OFFSET,swab_mac,6);
	xfer_riebl(init_mac,base+INIT_OFFSET,6);
	swab(init_mac,init_mac,6);
	if (memcmp(updt_mac,init_mac,6) != 0)
		error++;

	display_mac("New configured",config_mac);
	display_mac("New init block",init_mac);
	if (error)
		quit(1,"Update failed");

	quit(0,"");
	return 0;
}

/*
 *	display MAC address
 */
void display_mac(char *desc,char *mac)
{
	printf("%s MAC address is %02x:%02x:%02x:%02x:%02x:%02x\n",
			desc,mac[0],mac[1],mac[2],mac[3],mac[4],mac[5]);
}

/*
 *	exit with message
 */
void quit(int rc,char *msg)
{
	printf(msg);
	printf("\nPress a key to exit");
	getchar();
	exit(rc);
}

/*
 *	validate ascii MAC address & convert to 6-byte string
 */
int valid_mac(char *mac,char *s)
{
int i, n;
char *token;
long valid = MAC_PREFIX;

	for (i = 0; i < 6; i++) {
		token = strtok((i==0)?s:NULL,":.-");
		if (!token)
			return 0;
		if (strlen(token) != 2)
			return 0;
		if ((n=atox(token)) < 0)
			return 0;
		mac[i] = n;
	}

	if (memcmp(mac,(char *)&valid,4))
		return 0;

	return 1;
}

/*
 *	convert (assumed) hex ascii string
 */
int atox(char *hex)
{
int digit, n = 0;

	for ( ; *hex; hex++) {
		if ((*hex >= '0') && (*hex <= '9'))
			digit = *hex - '0';
		else if ((*hex >= 'A') && (*hex <= 'F'))
			digit = *hex - 'A' + 10;
		else if ((*hex >= 'a') && (*hex <= 'f'))
			digit = *hex - 'a' + 10;
		else return -1;
		n = (n << 4) + digit;
	}

	return n;
}

/*
 *	routines to transfer to/from Riebl card memory
 */
char *To;
char *From;
int Length;

void xfer_riebl(char *to,char *from,int length)
{
	To = to;
	From = from,
	Length = length;
	Supexec(copy);
}

int copy()
{
	memcpy(To,From,Length);
	return 0;
}

